# Copyright (C) 2017 Antony Chazapis SV9OAN
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import logging
import struct
import time
import Queue

from dstar import DSTARCallsign, DSTARModule
from stream import Packet, DVHeaderPacket, DVFramePacket
from network import UDPClientSocket
from utils import or_valueerror, StoppableThread

class DExtraConnectPacket(Packet):
    __slots__ = ['src_callsign', 'src_module', 'dest_module', 'revision']

    def __init__(self, src_callsign, src_module, dest_module, revision):
        self.src_callsign = src_callsign
        self.src_module = src_module
        self.dest_module = dest_module
        self.revision = revision

    @classmethod
    def from_data(cls, data):
        or_valueerror(len(data) == 11)
        src_callsign, src_module, dest_module, src_revision = struct.unpack('8sccB', data)
        src_callsign = DSTARCallsign(src_callsign)
        src_module = DSTARModule(src_module)
        dest_module = DSTARModule(dest_module)
        or_valueerror(str(dest_module) != ' ')
        if src_revision == 11:
            revision = 1
        elif (str(src_callsign).startswith('XRF')):
            revision = 2
        else:
            revision = 0
        return cls(src_callsign, src_module, dest_module, revision)

    def to_data(self):
        return struct.pack('8sccB', str(self.src_callsign),
                                    str(self.src_module),
                                    str(self.dest_module),
                                    11 if self.revision == 1 else 0) # XXX What about revisions 0 and 2?

class DExtraConnectAckPacket(Packet):
    __slots__ = ['src_callsign', 'src_module', 'dest_module', 'revision']

    def __init__(self, src_callsign, src_module, dest_module, revision):
        self.src_callsign = src_callsign
        self.src_module = src_module
        self.dest_module = dest_module
        self.revision = revision

    @classmethod
    def from_data(cls, data):
        if (len(data) == 11):
            raise NotImplemented

        or_valueerror(len(data) == 14)
        src_callsign, src_module, dest_module, ack = struct.unpack('8scc4s', data)
        src_callsign = DSTARCallsign(src_callsign)
        src_module = DSTARModule(src_module)
        dest_module = DSTARModule(dest_module)
        or_valueerror(str(dest_module) != ' ')
        or_valueerror(ack == 'ACK\x00')
        return cls(src_callsign, src_module, dest_module, 1) # XXX Could it be revision 0?

    @classmethod
    def from_connect_packet(cls, connect_packet):
        return cls(connect_packet.src_callsign,
                   connect_packet.src_module,
                   connect_packet.dest_module,
                   connect_packet.revision)

    def to_data(self):
        if self.revision == 2:
            return struct.pack('8sccB', str(self.src_callsign),
                                        str(self.dest_module),
                                        str(self.src_module),
                                        0)

        return struct.pack('8scc4s', str(self.src_callsign),
                                     str(self.src_module),
                                     str(self.dest_module),
                                     'ACK\x00')

class DExtraConnectNackPacket(Packet):
    __slots__ = ['src_callsign', 'src_module', 'dest_module']

    def __init__(self, src_callsign, src_module, dest_module):
        self.src_callsign = src_callsign
        self.src_callsign = src_module
        self.dest_module = dest_module

    @classmethod
    def from_data(cls, data):
        or_valueerror(len(data) == 14)
        src_callsign, src_module, dest_module, nack = struct.unpack('8scc4s', data)
        src_callsign = DSTARCallsign(src_callsign)
        src_module = DSTARModule(src_module)
        dest_module = DSTARModule(dest_module)
        or_valueerror(str(dest_module) != ' ')
        or_valueerror(nack == 'NAK\x00')
        return cls(src_callsign, src_module, dest_module)

    @classmethod
    def from_connect_packet(cls, connect_packet):
        return cls(connect_packet.src_callsign,
                   connect_packet.src_module,
                   connect_packet.dest_module)

    def to_data(self):
        return struct.pack('8scc4s', str(self.src_callsign),
                                     str(self.src_module),
                                     str(self.dest_module),
                                     'NAK\x00')

class DExtraDisconnectPacket(Packet):
    __slots__ = ['src_callsign', 'src_module']

    def __init__(self, src_callsign, src_module):
        self.src_callsign = src_callsign
        self.src_module = src_module

    @classmethod
    def from_data(cls, data):
        or_valueerror(len(data) == 11)
        src_callsign, src_module, dest_module, _ = struct.unpack('8sccB', data)
        src_callsign = DSTARCallsign(src_callsign)
        src_module = DSTARModule(src_module)
        dest_module = DSTARModule(dest_module)
        or_valueerror(str(dest_module) == ' ')
        return cls(src_callsign, src_module)

    def to_data(self):
        return struct.pack('8sc2s', str(self.src_callsign),
                                    str(self.src_module),
                                    '  ')

class DExtraDisconnectAckPacket(Packet):
    @classmethod
    def from_data(cls, data):
        or_valueerror(len(data) == 12)
        or_valueerror(data == 'DISCONNECTED')
        return cls()

    def to_data(self):
        return 'DISCONNECTED'

class DExtraKeepAlivePacket(Packet):
    __slots__ = ['src_callsign']

    def __init__(self, src_callsign):
        self.src_callsign = src_callsign

    @classmethod
    def from_data(cls, data):
        or_valueerror(len(data) == 9)
        src_callsign, _ = struct.unpack('8sc', data)
        src_callsign = DSTARCallsign(src_callsign)
        return cls(src_callsign) # XXX Get module?

    def to_data(self):
        return str(self.src_callsign) + '\x00' # XXX Send module?

class DExtraReceiveThread(StoppableThread):
    def __init__(self, callsign, sock, name='DExtraReceiveThread'):
        self.logger = logging.getLogger(self.__class__.__name__)

        StoppableThread.__init__(self, name=name)
        self._sleep_period = 0.01

        self.callsign = callsign
        self.sock = sock
        self.queue = Queue.Queue()

    def loop(self):
        while True: # While there is data to read from the socket
            data = self.sock.read()
            if not data:
                return

            try:
                packet = DVFramePacket.from_data(data)
            except ValueError:
                pass
            else:
                self.logger.debug('received dvframe packet from stream %s%s', packet.stream_id, ' (last)' if packet.is_last else '')
                self.queue.put(packet)
                continue

            try:
                packet = DVHeaderPacket.from_data(data)
            except ValueError:
                pass
            else:
                self.logger.debug('received dvheader packet from stream %s', packet.stream_id)
                self.queue.put(packet)
                continue

            try:
                packet = DExtraConnectAckPacket.from_data(data)
            except ValueError:
                pass
            else:
                self.logger.debug('received connect ack packet')
                self.queue.put(packet)
                continue

            try:
                packet = DExtraConnectNackPacket.from_data(data)
            except ValueError:
                pass
            else:
                self.logger.debug('received connect nack packet')
                self.queue.put(packet)
                continue

            try:
                packet = DExtraDisconnectPacket.from_data(data)
            except ValueError:
                pass
            else:
                self.logger.debug('received disconnect packet from %s', packet.src_callsign)
                self.queue.put(None) # Signal connection termination
                continue

            try:
                packet = DExtraDisconnectAckPacket.from_data(data)
            except ValueError:
                pass
            else:
                self.logger.debug('received disconnect ack packet')
                self.queue.put(packet)
                continue

            try:
                packet = DExtraKeepAlivePacket.from_data(data)
            except ValueError:
                pass
            else:
                # self.logger.debug('received keepalive packet from %s', packet.src_callsign)
                keepalive_packet = DExtraKeepAlivePacket(self.callsign)
                self.sock.write(keepalive_packet.to_data())
                continue

            self.logger.warning('unknown data received')

class DExtraDisconnectedError(Exception):
    pass

class DExtraConnection(object):
    DEFAULT_PORT = 30001

    def __init__(self, callsign, module, reflector_callsign, reflector_module, reflector_address):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.debug('initialized with callsign %s module %s reflector callsign %s reflector_module %s reflector_address %s', callsign, module, reflector_callsign, reflector_module, reflector_address)

        self.callsign = callsign
        self.module = module
        self.reflector_callsign = reflector_callsign
        self.reflector_module = reflector_module
        self.reflector_address = reflector_address

        self.sock = None
        self.receive_thread = None
        self.disconnected = False

    def _connect(self, timeout=3):
        connect_packet = DExtraConnectPacket(self.callsign, self.module, self.reflector_module, 1)
        self.sock.write(connect_packet.to_data())

        clock = time.time()
        limit = clock + timeout
        while (clock < limit):
            packet = self.read(timeout=limit - clock)
            if packet:
                if isinstance(packet, DExtraConnectAckPacket):
                    return True
                if isinstance(packet, DExtraConnectNackPacket):
                    return False
            clock = time.time()
        return False

    def _disconnect(self, timeout=3):
        disconnect_packet = DExtraDisconnectPacket(self.callsign, self.module)
        self.sock.write(disconnect_packet.to_data())

        clock = time.time()
        limit = clock + timeout
        while (clock < limit):
            packet = self.read(timeout=limit - clock)
            if packet and isinstance(packet, DExtraDisconnectAckPacket):
                return True
            clock = time.time()
        return False

    def open(self):
        or_valueerror(self.sock is None)

        try:
            self.sock = UDPClientSocket(self.reflector_address)
            self.sock.open()
        except Exception as e:
            self.logger.error('can not open UDP socket: %s', str(e))
            self.sock = None
            return False
        self.logger.info('connected to reflector %s at address %s', self.reflector_callsign, self.reflector_address)

        self.receive_thread = DExtraReceiveThread(self.callsign, self.sock)
        self.receive_thread.start()
        self.disconnected = False

        return self._connect()

    def close(self):
        while not self.receive_thread.queue.empty():
            self.receive_thread.queue.get()

        if not self.disconnected:
            self._disconnect()

        self.receive_thread.join()

        if self.sock:
            self.sock.close()
        self.sock = None
        self.logger.info('disconnected from reflector %s at address %s', self.reflector_callsign, self.reflector_address)

    def __enter__(self):
        if not self.open():
            raise Exception('can not open DExtra connection to %s' % (self.reflector_address,))
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def read(self, timeout=3):
        elapsed = 0
        step = 0.01
        while elapsed < timeout:
            try:
                packet = self.receive_thread.queue.get(True, step)
                if not packet:
                    self.disconnected = True
                    raise DExtraDisconnectedError
                return packet
            except Queue.Empty:
                time.sleep(step)
                elapsed += step
                continue
        return None

    def write(self, packet):
        return self.sock.write(packet.to_data())



# Copyright (C) 2018 Antony Chazapis SV9OAN
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import struct

from dstar import DSTARHeader, DSTARFrame
from utils import or_valueerror

class Packet(object):
    pass

class DVHeaderPacket(Packet):
    __slots__ = ['band_1', 'band_2', 'band_3', 'stream_id', 'dstar_header']

    def __init__(self, band_1, band_2, band_3, stream_id, dstar_header):
        self.band_1 = band_1
        self.band_2 = band_2
        self.band_3 = band_3
        self.stream_id = stream_id
        self.dstar_header = dstar_header

    @classmethod
    def from_data(cls, data):
        or_valueerror(len(data) == 56)
        or_valueerror(data[:4] == 'DSVT')
        or_valueerror(data[4] == '\x10')
        or_valueerror(data[8] == '\x20')
        band_1, band_2, band_3, stream_id = struct.unpack('<BBBH', data[9:14])
        dstar_header = DSTARHeader.from_data(data[15:])
        return cls(band_1, band_2, band_3, stream_id, dstar_header)

    def to_data(self):
        return ('DSVT\x10\x00\x00\x00\x20' +
                struct.pack('<BBBH', self.band_1, self.band_2, self.band_3, self.stream_id) +
                '\x80' +
                self.dstar_header.to_data())

class DVFramePacket(Packet):
    __slots__ = ['band_1', 'band_2', 'band_3', 'stream_id', 'packet_id', 'dstar_frame']

    def __init__(self, band_1, band_2, band_3, stream_id, packet_id, dstar_frame):
        self.band_1 = band_1
        self.band_2 = band_2
        self.band_3 = band_3
        self.stream_id = stream_id
        self.packet_id = packet_id
        self.dstar_frame = dstar_frame

    @property
    def is_last(self):
        return self.packet_id & 64 != 0

    @classmethod
    def from_data(cls, data):
        or_valueerror(len(data) == 27)
        or_valueerror(data[:4] == 'DSVT')
        or_valueerror(data[4] == '\x20')
        or_valueerror(data[8] == '\x20')
        band_1, band_2, band_3, stream_id, packet_id = struct.unpack('<BBBHB', data[9:15])
        dstar_frame = DSTARFrame.from_data(data[15:])
        return cls(band_1, band_2, band_3, stream_id, packet_id, dstar_frame)

    def to_data(self):
        return ('DSVT\x20\x00\x00\x00\x20' +
                struct.pack('<BBBHB', self.band_1, self.band_2, self.band_3, self.stream_id, self.packet_id) +
                self.dstar_frame.to_data())



# Copyright (C) 2017 Antony Chazapis SV9OAN
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import struct
import string

from crc import CCITTChecksum
from utils import or_valueerror

class DSTARCallsign(object):
    __slots__ = ['callsign']

    def __init__(self, callsign):
        or_valueerror(len(callsign) > 3)
        or_valueerror(set(callsign[:3]).issubset(string.ascii_letters + string.digits))
        or_valueerror(not callsign[:3].isdigit())
        or_valueerror(set(callsign[3:]).issubset(string.ascii_letters + string.digits + ' '))
        or_valueerror(len(callsign) <= 8)

        self.callsign = callsign.upper()

    def __str__(self):
        return '{: <8}'.format(self.callsign)

class DSTARSuffix(object):
    __slots__ = ['suffix']

    def __init__(self, suffix):
        or_valueerror(set(suffix).issubset(string.ascii_letters + string.digits + ' '))
        or_valueerror(len(suffix) <= 4)

        self.suffix = suffix.upper()

    def __str__(self):
        return '{: <4}'.format(self.suffix)

class DSTARModule(object):
    __slots__ = ['module']

    def __init__(self, module):
        or_valueerror(len(module) == 1)
        or_valueerror(set(module).issubset(string.ascii_letters + ' '))

        self.module = module.upper()

    def __str__(self):
        return self.module

class DSTARHeader(object):
    __slots__ = ['flag_1',
                 'flag_2',
                 'flag_3',
                 'repeater_1_callsign',
                 'repeater_2_callsign',
                 'ur_callsign',
                 'my_callsign',
                 'my_suffix']

    def __init__(self,
                 flag_1,
                 flag_2,
                 flag_3,
                 repeater_1_callsign,
                 repeater_2_callsign,
                 ur_callsign,
                 my_callsign,
                 my_suffix):
        self.flag_1 = flag_1
        self.flag_2 = flag_2
        self.flag_3 = flag_3
        self.repeater_1_callsign = repeater_1_callsign
        self.repeater_2_callsign = repeater_2_callsign
        self.ur_callsign = ur_callsign
        self.my_callsign = my_callsign
        self.my_suffix = my_suffix

    @classmethod
    def from_data(cls, data):
        or_valueerror(len(data) == 41)
        header, crc = data[:39], data[39:]
        flag_1, \
        flag_2, \
        flag_3, \
        repeater_1_callsign, \
        repeater_2_callsign, \
        ur_callsign, \
        my_callsign, \
        my_suffix = struct.unpack('BBB8s8s8s8s4s', header)
        repeater_1_callsign = DSTARCallsign(repeater_1_callsign)
        repeater_2_callsign = DSTARCallsign(repeater_2_callsign)
        ur_callsign = DSTARCallsign(ur_callsign)
        my_callsign = DSTARCallsign(my_callsign)
        my_suffix = DSTARSuffix(my_suffix)
        checksum = CCITTChecksum()
        checksum.update(header)
        or_valueerror(crc == checksum.result())
        return cls(flag_1,
                   flag_2,
                   flag_3,
                   repeater_1_callsign,
                   repeater_2_callsign,
                   ur_callsign,
                   my_callsign,
                   my_suffix)

    def to_data(self):
        header = struct.pack('BBB8s8s8s8s4s', self.flag_1,
                                              self.flag_2,
                                              self.flag_3,
                                              str(self.repeater_1_callsign),
                                              str(self.repeater_2_callsign),
                                              str(self.ur_callsign),
                                              str(self.my_callsign),
                                              str(self.my_suffix))
        checksum = CCITTChecksum()
        checksum.update(header)
        crc = checksum.result()
        return header + crc

class DSTARFrame(object):
    __slots__ = ['dvcodec', 'dvdata']

    def __init__(self, dvcodec, dvdata):
        self.dvcodec = dvcodec
        self.dvdata = dvdata

    @classmethod
    def from_data(cls, data):
        or_valueerror(len(data) == 12)
        dvcodec, dvdata = data[:9], data[9:]
        return cls(dvcodec, dvdata)

    def to_data(self):
        return self.dvcodec + ('\x00' * (9 - len(self.dvcodec))) + \
               self.dvdata + ('\x00' * (3 - len(self.dvdata)))



# Copyright (C) 2018 Antony Chazapis SV9OAN
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import os
import sys
import argparse
import logging
import wave
import struct

import pydv.mbelib
import pydv.codec2

from stream import DVHeaderPacket, DVFramePacket
from dvtool import DVToolFile

def dv_decoder():
    parser = argparse.ArgumentParser(description='D-STAR decoder. Decodes recordings.')
    parser.add_argument('-v', '--verbose', default=False, action='store_true', help='enable debug output')
    parser.add_argument('input', help='name of file to decode (DVTool format)')
    parser.add_argument('output', help='name of file to write (WAV format)')
    args = parser.parse_args()

    logging.basicConfig(format='%(asctime)s [%(levelname)7s] %(name)s: %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S',
                        level=logging.DEBUG if args.verbose else logging.INFO)
    logger = logging.getLogger(os.path.basename(sys.argv[0]))

    try:
        with DVToolFile(args.input) as f:
            stream = f.read()
    except Exception as e:
        raise
        logger.error(str(e))
        sys.exit(1)

    header = stream[0]
    if not isinstance(header, DVHeaderPacket):
        logger.error('first packet in stream is not a header')
        sys.exit(1)

    # Determine vocoder (SV9OAN extension)
    version = header.dstar_header.flag_3 & 0xff
    if version == 0:
        vocoder = 'ambe'
        logger.info('stream encoded with AMBE vocoder')
    elif version & 0x01 == 0x01:
        # The first bit controls the vocoder type:
        #   0: AMBE (backwards compatible)
        #   1: Codec 2
        #
        # The second bit differentiates between modes, while the third
        # enables FEC (not implemented):
        # This results in the following combinations of bits 2 and 3:
        #   00: 3200 mode (160 samples/20 ms into 64 bits)
        #   01: 2400 mode (160 samples/20 ms into 48 bits)
        #   10: 3200 mode (160 samples/20 ms into 64 bits + ? bits FEC)
        #   11: 2400 mode (160 samples/20 ms into 48 bits + ? bits FEC)
        #
        # The space available in the frame is 72 bits, so that leaves
        # us with 8 bits in the case of 3200 mode (is it enough?) and
        # 24 in the case of 2400 mode. The latter is enough for 22 bits
        # of FEC, as FreeDV does in 2400 and 1850 modes.
        vocoder = 'codec2'
        mode = pydv.codec2.MODE_2400 if (version & 0x02 == 0x02) else pydv.codec2.MODE_3200
        fec = True if (version & 0x04 == 0x04) else False # Not implemented
        logger.info('stream encoded with Codec 2 vocoder (mode %s, fec: %s)',
                    '2400' if mode == pydv.codec2.MODE_2400 else '3200',
                    'on' if fec else 'off')

        if fec:
            logger.error('FEC is not implemented')
            sys.exit(1)

    wavef = wave.open(args.output,'w')
    wavef.setnchannels(1)
    wavef.setsampwidth(2)
    wavef.setframerate(8000)

    if vocoder == 'ambe':
        state = pydv.mbelib.init_state()
        for packet in stream:
            if not isinstance(packet, DVFramePacket):
                continue
            samples = pydv.mbelib.decode_dstar(state, packet.dstar_frame.dvcodec)
            data = struct.pack('<160h', *samples)
            wavef.writeframes(data)
    else:
        state = pydv.codec2.init_state(mode)
        for packet in stream:
            if not isinstance(packet, DVFramePacket):
                continue
            dvcodec = packet.dstar_frame.dvcodec[:6] if mode == pydv.codec2.MODE_2400 else packet.dstar_frame.dvcodec[:8]
            data = pydv.codec2.decode(state, dvcodec)
            wavef.writeframes(data)

    wavef.close()
    logger.info('output written to %s' % args.output)

def main():
    dv_decoder()

if __name__ == '__main__':
    dv_decoder()



# Copyright (C) 2018 Antony Chazapis SV9OAN
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import os
import sys
import argparse
import logging

from dstar import DSTARCallsign, DSTARModule
from dextra import DExtraConnection, DExtraDisconnectedError
from stream import DVHeaderPacket, DVFramePacket
from network import NetworkAddress
from dvtool import DVToolFile

def dv_recorder():
    parser = argparse.ArgumentParser(description='D-STAR recorder. Connects to reflector and records traffic.')
    parser.add_argument('-v', '--verbose', default=False, action='store_true', help='enable debug output')
    parser.add_argument('callsign', help='your callsign')
    parser.add_argument('reflector', help='reflector\'s callsign')
    parser.add_argument('module', help='reflector\'s module')
    parser.add_argument('address', help='reflector\'s hostname or IP address')
    args = parser.parse_args()

    logging.basicConfig(format='%(asctime)s [%(levelname)7s] %(name)s: %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S',
                        level=logging.DEBUG if args.verbose else logging.INFO)
    logger = logging.getLogger(os.path.basename(sys.argv[0]))

    try:
        callsign = DSTARCallsign(args.callsign)
        reflector_callsign = DSTARCallsign(args.reflector)
        reflector_module = DSTARModule(args.module)
        reflector_address = NetworkAddress(args.address, DExtraConnection.DEFAULT_PORT)
    except ValueError:
        print parser.print_help()
        sys.exit(1)

    try:
        stream_id = None
        with DExtraConnection(callsign, DSTARModule(' '), reflector_callsign, reflector_module, reflector_address) as conn:
            try:
                while True:
                    packet = conn.read()
                    if isinstance(packet, DVHeaderPacket):
                        if packet.stream_id != stream_id:
                            stream_id = packet.stream_id
                            stream = [packet]
                    elif isinstance(packet, DVFramePacket):
                        if packet.stream_id == stream_id:
                            stream.append(packet)
                            if packet.is_last:
                                with DVToolFile('%s.dvtool' % stream_id) as f:
                                    f.write(stream)
                                stream_id = None
                    else:
                        pass
            except (DExtraDisconnectedError, KeyboardInterrupt):
                pass
    except Exception as e:
        logger.error(str(e))
        sys.exit(1)

def main():
    dv_recorder()

if __name__ == '__main__':
    dv_recorder()



# Copyright (C) 2018 Antony Chazapis SV9OAN
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import os
import sys
import argparse
import logging
import wave
import struct

import pydv.mbelib
import pydv.codec2

from stream import DVHeaderPacket, DVFramePacket
from dvtool import DVToolFile

def dv_transcoder():
    parser = argparse.ArgumentParser(description='D-STAR transcoder. Transcodes recordings from AMBE to Codec 2.')
    parser.add_argument('-v', '--verbose', default=False, action='store_true', help='enable debug output')
    parser.add_argument('-m', '--mode', default='3200', help='Codec 2 mode (2400 or 3200)')
    parser.add_argument('-f', '--fec', default=False, action='store_true', help='enable FEC')
    parser.add_argument('input', help='name of file to transcode (DVTool format)')
    parser.add_argument('output', help='name of file to write (DVTool format)')
    args = parser.parse_args()

    logging.basicConfig(format='%(asctime)s [%(levelname)7s] %(name)s: %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S',
                        level=logging.DEBUG if args.verbose else logging.INFO)
    logger = logging.getLogger(os.path.basename(sys.argv[0]))

    if args.mode not in ('2400', '3200'):
        logger.error('mode can be either 2400 or 3200')
        sys.exit(1)
    if args.fec:
        logger.error('FEC is not implemented')
        sys.exit(1)

    try:
        with DVToolFile(args.input) as f:
            stream = f.read()
    except Exception as e:
        logger.error(str(e))
        sys.exit(1)

    header = stream[0]
    if not isinstance(header, DVHeaderPacket):
        logger.error('first packet in stream is not a header')
        sys.exit(1)

    # Determine vocoder and set (SV9OAN extension)
    version = header.dstar_header.flag_3 & 0xff
    if version & 0x01 == 0x01:
        logger.info('stream already encoded with Codec 2')
        sys.exit(0)
    
    version = 0x01
    if args.mode == '2400':
        version |= 0x02
    # if args.fec:
    #     version &= 0x04
    header.dstar_header.flag_3 = version

    state_mbelib = pydv.mbelib.init_state()
    mode = pydv.codec2.MODE_2400 if (args.mode == '2400') else pydv.codec2.MODE_3200
    state_codec2 = pydv.codec2.init_state(mode)
    for packet in stream:
        if not isinstance(packet, DVFramePacket):
            continue
        samples = pydv.mbelib.decode_dstar(state_mbelib, packet.dstar_frame.dvcodec)
        data = struct.pack('<160h', *samples)
        packet.dstar_frame.dvcodec = pydv.codec2.encode(state_codec2, data)

    with DVToolFile(args.output) as f:
        f.write(stream)

def main():
    dv_transcoder()

if __name__ == '__main__':
    dv_transcoder()
